!= ȳɹ
!
!= Setting constants of snow and sea ice
!
! Authors::   Yoshiyuki O. Takahashi
! Version::   $Id: constants_snowseaice.f90,v 1.3 2012-01-20 00:18:56 yot Exp $ 
! Tag Name::  $Name: dcpam5-20130930 $
! Copyright:: Copyright (C) GFD Dennou Club, 2008. All rights reserved.
! License::   See COPYRIGHT[link:../../../COPYRIGHT]
!

module constants_snowseaice
  !
  != ȳɹ
  !
  != Setting constants of snow and sea ice
  !
  ! <b>Note that Japanese and English are described in parallel.</b>
  !
  ! ȳɹꤪݴɤԤޤ. 
  ! ͤ NAMELIST ѿ constants_snowseaice_nml ˤäѹ뤳Ȥ
  ! ǽǤ. 
  !
  ! Constants of snow and sea ice are set and stored. 
  ! These values can be changed by NAMELIST group name 
  ! "constants_snowseaice_nml". 
  !
  !== Procedures List
  !
  ! ConstantsSnowSeaIceInit :: ʪ
  ! ------------  :: ------------
  ! ConstantsSnowSeaIceInit :: Settings of constants of snow and sea ice
  !
  !== NAMELIST
  !
  ! NAMELIST#constants_snowseaice_nml
  !


  ! ⥸塼 ; USE statements
  !

  ! ̷ѥ᥿
  ! Kind type parameter
  !
  use dc_types, only: DP     ! ټ¿. Double precision. 

  ! ʸ ; Declaration statements
  !
  implicit none
  private

  ! ³
  ! Public procedure
  !
  public:: ConstantsSnowSeaIceInit

  ! ѿ
  ! Public variables
  !
  logical, save :: constants_snowseaice_inited = .false.
                              ! ե饰. 
                              ! Initialization flag

  real(DP), public, save:: ThresholdSurfSnow
  real(DP), public, save:: SnowAlbedo
  real(DP), public, save:: TempCondWater
                        ! <Japanese>
                        ! Condensation temperature of water

  real(DP), public, save:: SeaIceVolHeatCap
                              ! Volumetric heat capacity of sea ice (J m-3 K-1)
  real(DP), public, save:: SeaIceThermCondCoef
                              ! Thermal conductivity of sea ice (W m-1 K-1)

  real(DP), public, save:: SeaIceThreshold
  real(DP), public, save:: SeaIceThickness
  real(DP), public, save:: TempBelowSeaIce

  real(DP), public, save:: SeaIceAlbedo


  real(DP), public, save:: CO2IceThreshold
  real(DP), public, save:: CO2IceAlbedoS
  real(DP), public, save:: CO2IceAlbedoN
  real(DP), public, save:: CO2IceEmisS
  real(DP), public, save:: CO2IceEmisN



  ! ѿ
  ! Private variables
  !

  character(*), parameter:: module_name = 'constants_snowseaice'
                              ! ⥸塼̾. 
                              ! Module name
  character(*), parameter:: version = &
    & '$Name: dcpam5-20130930 $' // &
    & '$Id: constants_snowseaice.f90,v 1.3 2012-01-20 00:18:56 yot Exp $'
                              ! ⥸塼ΥС
                              ! Module version

contains

  !--------------------------------------------------------------------------------------

  subroutine ConstantsSnowSeaIceInit
    !
    ! constants_snowseaice ⥸塼νԤޤ. 
    ! NAMELIST#constants_snowseaice_nml ɤ߹ߤϤμ³ǹԤޤ. 
    !
    ! "constants_snowseaice" module is initialized. 
    ! NAMELIST#constants_snowseaice_nml is loaded in this procedure. 
    !

    ! ⥸塼 ; USE statements
    !

    ! NAMELIST եϤ˴ؤ桼ƥƥ
    ! Utilities for NAMELIST file input
    !
    use namelist_util, only: namelist_filename, NmlutilMsg

    ! ե
    ! File I/O support
    !
    use dc_iounit, only: FileOpen

    ! ̷ѥ᥿
    ! Kind type parameter
    !
    use dc_types, only: STDOUT ! ɸϤֹ. Unit number of standard output

    ! å
    ! Message output
    !
    use dc_message, only: MessageNotify

    ! ʸ ; Declaration statements
    !
    implicit none

    integer:: unit_nml        ! NAMELIST ե륪ץֹ. 
                              ! Unit number for NAMELIST file open
    integer:: iostat_nml      ! NAMELIST ɤ߹߻ IOSTAT. 
                              ! IOSTAT of NAMELIST read

    ! NAMELIST ѿ
    ! NAMELIST group name
    !
    namelist /constants_snowseaice_nml/ &
      & ThresholdSurfSnow  , &
      & SnowAlbedo         , &
      & TempCondWater      , &
      & SeaIceVolHeatCap   , &
      & SeaIceThermCondCoef, &
      & SeaIceThreshold    , &
      & SeaIceThickness    , &
      & TempBelowSeaIce    , &
      & SeaIceAlbedo       , &
      & CO2IceAlbedoS      , &
      & CO2IceAlbedoN      , &
      & CO2IceEmisS        , &
      & CO2IceEmisN

          !
          ! ǥեͤˤĤƤϽ³ "constants#ConstantsInit" 
          ! Υɤ򻲾ȤΤ. 
          !
          ! Refer to source codes in the initialization procedure
          ! "constants#ConstantsInit" for the default values. 
          !

    ! ¹ʸ ; Executable statement
    !

    if ( constants_snowseaice_inited ) return


    ! ǥեͤ
    ! Default values settings
    !

    ThresholdSurfSnow  = 0.1d0 * 1.0d3 * 0.01d0
                        ! 1 cm snow depth
                        ! 0.1 * 1000 (kg m-3; water density) x 1 cm
                        ! This value is selected arbitrarily. 
!!$    SnowAlbedo         = 0.7d0
    SnowAlbedo         = 0.75d0
                        ! This value is selected arbitrarily. 
    TempCondWater      = 273.15d0
                        ! <Japanese>
                        ! Condensation temperature of water

!!$    ! https://www.myroms.org/wiki/index.php/Sea-Ice_Model
!!$    SeaIceDens         = 900.0d0   ! no reliable reference
!!$    SeaIceSpecHeat     = 2093.0d0  ! no reliable reference
!!$    SeaIceHeatDiffCoef = 2.04d0 / ( SeaIceDens * SeaIceSpecHeat )  ! 2.04d0 (W (m K)-1),  no reliable reference

!!$    SeaIceDens         = 0.92d3
!!$    ! kg m-3
!!$    SeaIceHeatDiffCoef = 2.04d0 / ( SeaIceDens * SeaIceSpecHeat )
!!$    ! 2.04d0 (W (m K)-1),  no reliable reference
!!$    SeaIceSpecHeat     = 1.9d6 / SeaIceDens
!!$    ! no reliable reference

    SeaIceVolHeatCap    = 1.9d6
    ! J m-3 K-1
    ! Value of ice in Table 12.1 by Hillel (2004). 
    SeaIceThermCondCoef = 2.2d0
    ! W m-1 K-1
    ! Value of ice in Table 12.2 by Hillel (2004). 

    ! Reference
    !
    ! Hillet, D., 
    !   Introduction to Environmental Soil Physics, 
    !   Elsevier Academic Press, pp494, 2004.


    SeaIceThreshold    = 0.5d0            ! arbitrarily set, no reliable reference
    SeaIceThickness    = 2.0d0            ! arbitrarily set, no reliable reference
    TempBelowSeaIce    = 273.15d0 - 2.0d0 ! arbitrarily set, no reliable reference

!!$    SeaIceAlbedo       = 0.5d0            ! arbitrarily set, no reliable reference
    SeaIceAlbedo       = 0.75d0           ! arbitrarily set, no reliable reference


    CO2IceThreshold = 1.0_DP         ! No reference
    CO2IceAlbedoS   = 0.75_DP        ! No reference
    CO2IceAlbedoN   = 0.75_DP        ! No reference
    CO2IceEmisS     = 1.0_DP         ! No reference
    CO2IceEmisN     = 1.0_DP         ! No reference


    ! NAMELIST 
    ! Input from NAMELIST
    !
    if ( trim(namelist_filename) /= '' ) then
      call FileOpen( unit_nml, &          ! (out)
        & namelist_filename, mode = 'r' ) ! (in)

      rewind( unit_nml )
      read( unit_nml,                     &  ! (in)
        & nml = constants_snowseaice_nml, &  ! (out)
        & iostat = iostat_nml            )   ! (out)
      close( unit_nml )

      call NmlutilMsg( iostat_nml, module_name ) ! (in)
      if ( iostat_nml == 0 ) write( STDOUT, nml = constants_snowseaice_nml )
    end if

    !  ; Print
    !
    call MessageNotify( 'M', module_name, '----- Initialization Messages -----' )
    call MessageNotify( 'M', module_name, '  ThresholdSurfSnow   = %f', d = (/ ThresholdSurfSnow   /) )
    call MessageNotify( 'M', module_name, '  SnowAlbedo          = %f', d = (/ SnowAlbedo          /) )
    call MessageNotify( 'M', module_name, '  TempCondWater       = %f', d = (/ TempCondWater       /) )
    call MessageNotify( 'M', module_name, '  SeaIceThreshold     = %f', d = (/ SeaIceThreshold     /) )
    call MessageNotify( 'M', module_name, '  SeaIceVolHeatCap    = %f', d = (/ SeaIceVolHeatCap    /) )
    call MessageNotify( 'M', module_name, '  SeaIceThermCondCoef = %f', d = (/ SeaIceThermCondCoef /) )

!!$    call MessageNotify( 'M', module_name, '  SeaIceDens         = %f', d = (/ SeaIceDens         /) )
!!$    call MessageNotify( 'M', module_name, '  SeaIceSpecHeat     = %f', d = (/ SeaIceSpecHeat     /) )
!!$    call MessageNotify( 'M', module_name, '  SeaIceHeatDiffCoef = %f', d = (/ SeaIceHeatDiffCoef /) )

    call MessageNotify( 'M', module_name, '  SeaIceThickness     = %f', d = (/ SeaIceThickness     /) )
    call MessageNotify( 'M', module_name, '  TempBelowSeaIce     = %f', d = (/ TempBelowSeaIce     /) )
    call MessageNotify( 'M', module_name, '  SeaIceAlbedo        = %f', d = (/ SeaIceAlbedo        /) )
    call MessageNotify( 'M', module_name, '  CO2IceAlbedoS       = %f', d = (/ CO2IceAlbedoS       /) )
    call MessageNotify( 'M', module_name, '  CO2IceAlbedoN       = %f', d = (/ CO2IceAlbedoN       /) )
    call MessageNotify( 'M', module_name, '  CO2IceEmisS         = %f', d = (/ CO2IceEmisS       /) )
    call MessageNotify( 'M', module_name, '  CO2IceEmisN         = %f', d = (/ CO2IceEmisN       /) )
    call MessageNotify( 'M', module_name, '-- version = %c', c1 = trim(version) )

    constants_snowseaice_inited = .true.

  end subroutine ConstantsSnowSeaIceInit

  !--------------------------------------------------------------------------------------

end module constants_snowseaice
